#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "lib/safem.h"
#include "lib/log.h"
#include "wad/wadio.h"
#include "animtex.h"

static const AnimTexTable default_att[]={
     {3,0,"NUKAGE1",1},
     {3,1,"NUKAGE2",1},
     {3,2,"NUKAGE3",1},
     {3,0,"BLOOD1",1},
     {3,1,"BLOOD2",1},
     {3,2,"BLOOD3",1},
     {4,0,"SWATER1",1},
     {4,1,"SWATER2",1},
     {4,2,"SWATER3",1},
     {4,3,"SWATER4",1},
     {4,0,"FWATER1",1},
     {4,1,"FWATER2",1},
     {4,2,"FWATER3",1},
     {4,3,"FWATER4",1},
     {6,0,"LAVA1",1},
     {6,1,"LAVA2",1},
     {6,2,"LAVA3",1},
     {6,3,"LAVA4",1},
     {6,4,"LAVA3",1},
     {6,5,"LAVA2",1},
   
     {4,0,"RROCK05",1},
     {4,1,"RROCK06",1},
     {4,2,"RROCK07",1},
     {4,3,"RROCK08",1},
     {4,0,"SLIME01",1},
     {4,1,"SLIME02",1},
     {4,2,"SLIME03",1},
     {4,3,"SLIME04",1},
     {4,0,"SLIME05",1},
     {4,1,"SLIME06",1},
     {4,2,"SLIME07",1},
     {4,3,"SLIME08",1},
     {4,0,"SLIME09",1},
     {4,1,"SLIME10",1},
     {4,2,"SLIME11",1},
     {4,3,"SLIME12",1},

     {4,0,"BFALL1",0},
     {4,1,"BFALL2",0},
     {4,2,"BFALL3",0},
     {4,3,"BFALL4",0},
     {4,0,"SFALL1",0},
     {4,1,"SFALL2",0},
     {4,2,"SFALL3",0},
     {4,3,"SFALL4",0},
     {4,0,"WFALL1",0},
     {4,1,"WFALL2",0},
     {4,2,"WFALL3",0},
     {4,3,"WFALL4",0},
     {4,0,"DBRAIN1",0},
     {4,1,"DBRAIN2",0},
     {4,2,"DBRAIN3",0},
     {4,3,"DBRAIN4",0},

     {4,0,"BLODGR1",0},
     {4,1,"BLODGR2",0},
     {4,2,"BLODGR3",0},
     {4,3,"BLODGR4",0},
     {4,0,"BLODRIP1",0},
     {4,1,"BLODRIP2",0},
     {4,2,"BLODRIP3",0},
     {4,3,"BLODRIP4",0},
     {2,0,"FIREBLU1",0},
     {2,1,"FIREBLU2",0},
     {2,0,"FIRELAV3",0},
     {2,1,"FIRELAVA",0},
     {3,0,"FIREMAG1",0},
     {3,1,"FIREMAG2",0},
     {3,2,"FIREMAG3",0},
     {3,0,"GSTFONT1",0},
     {3,1,"GSTFONT2",0},
     {3,2,"GSTFONT3",0},
     {3,0,"ROCKRED1",0},
     {3,1,"ROCKRED2",0},
     {3,2,"ROCKRED3",0},
     {3,0,"SLADRIP1",0},
     {3,1,"SLADRIP2",0},
     {3,2,"SLADRIP3",0},
   
     {0,0,"",0}
};

static LumpNum animtex_ln=BAD_LUMPNUM;
static int natt=0;
static const AnimTexTable *att=NULL;
static Texture **attd=NULL;

AnimTexNum get_animtex(const char *name) {
   int i;
   if(!att) init_animtex();
   for(i=0;i<natt;i++)  {
      if(!strncasecmp(name,att[i].name,8)) {
	 /* if my seqnum is bogus, return the next entry */
	 if(att[i].myseqnum>=att[i].seqlen) return i+1;
	 /* otherwise, I'll do */
	 else return i;
      };
      /* bogus seqnums indicate this is a name for the next set
       * of anims, whose individual names should be ignored
       */
      if(att[i].myseqnum>=att[i].seqlen) i+=att[i].seqlen;
     };
   return -1;
};

Texture *get_anim_texture(AnimTexNum num,unsigned int seqnum) {
   if(!attd) init_animtex();
   if(num>=natt||num<0)
     logprintf(LOG_FATAL,'A',"Bad AnimTexNum (%d)",num);
   num-=att[num].myseqnum;
   seqnum%=att[num].seqlen;
   
   num+=seqnum;
   if(attd[num]==NULL)  {
      if(att[num].isflat) attd[num]=get_flat_texture(att[num].name);
      else attd[num]=get_wall_texture(att[num].name);
   };
   return attd[num];
};

void init_animtex(void) {
   if(att) reset_animtex();
   animtex_ln=lookup_lump("ANIMTEX",NULL,NULL);
   if(LUMPNUM_OK(animtex_ln))  {
      natt=get_lump_len(animtex_ln)/sizeof(AnimTexTable);
      att=load_lump(animtex_ln);
   }
   else  {
      att=default_att;
      natt=0;
      while(att[natt].name[0]) natt++;
   };
   attd=safe_calloc(sizeof(Texture *),natt);
};

void reset_animtex(void) {
   if(LUMPNUM_OK(animtex_ln)) free_lump(animtex_ln);
   if(attd) safe_free(attd);
   att=NULL;
   attd=NULL;
   natt=0;
   animtex_ln=BAD_LUMPNUM;
};

