#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "lib/log.h"
#include "wad/wadio.h"
#include "texture.h"
#include "gettable.h"
#include "plat/video.h"
#include "render/render.h"
#include "render/draw.h"
#include "banner.h"
#include "things.h"
#include "updmap.h"
#include "plat/input.h"
#include "plat/sound.h"

#define DUMB "DUMB"
#define VERSION "0.04"

#ifdef __cplusplus
#define BANNER (DUMB "++ " VERSION)
#else
#define BANNER (DUMB " " VERSION)
#endif

static void usage(const char *name)  {
   printf("%s\nUsage: %s [options]\n"
	  "Where options are:\n"
	  "\t-q\t\tQuiet: don't log to the screen.\n"
	  "\t-x\t\tRelax sprite and flat searching.\n"
	  "\t-n\t\tNo Objects: don't create any objects, just preview map.\n"
	  "\t-w <wadfile> [wadfile] [wadfile]...\n"
	  "\t-l <logfile> [logfile] [logfile]...\n"
	  "\t-r <fname>\t\tRecord player's actions to <fname>\n"
	  "\t-p <fname>\t\tPlayback player's actions from <fname>\n"
	  "\t-m <mapname>\n"
	  "\t-d <width> <height>\n"
	  "\t-b <bpp>\n"
	  "\t-v <n>\t\tViewport will take up (n-1)/n of screen.\n"
	  "\t-h <num>\t\tSet difficulty level to <num>\n"
	  ,BANNER,name);
   exit(1);
};

#define usage() usage(argv[0])

#define MAX_WADS 16

/* this should get moved into game.c */
static int width=320,height=200,bpp=1;
static int banner=-1,wbanner=-1,bfont=-1;
void gmsg(const char *s) {
   add_to_banner(banner,NULL,width,0);
   add_str_to_banner(banner,bfont,s);
};

int main(int argc,char **argv) {
   int nwads=0;
   const char *wadf[MAX_WADS];
   int vpdiv=0;
   void *fb;
   View view;
   LevData ld[1];
   /*int i;
   Texture *t;*/
   ThingDyn *td;
   const SectorDyn *sd;
   int tt,frames=0;
   int i,quiet=0,no_objs=0;
   int follow,shframes=0;
   const char *mapname="E1M1";
   FILE *frec=NULL,*fplay=NULL;
   const char *testarg=NULL;
   Texture *tdraw=NULL;
   int difficulty=3;

   LumpNum sound_ln=BAD_LUMPNUM;
   const unsigned char *sound=NULL;

   INIT_RENDERER_SPEC((*init_renderer))=NULL;
   RENDER_SPEC((*render))=NULL;
   
   video_preinit();

   for(i=1;i<argc;i++)  {
      char c;
      if(argv[i][0]!='-') usage();
      switch(c=argv[i][1])  {
       /* q = quiet */
       case('q'):
	 quiet=1;
	 break;
       /* n = no objects */
       case('n'):
	 no_objs=1;
	 break;
       /* x = relax */
       case('x'):
	 texture_flags|=TF_RELAX_SPRITES|TF_RELAX_FLATS;
	 break;
       /* w = wad */
       case('w'):
	 while(++i<argc) {
	    if(argv[i][0]=='-')  {
	       i--;
	       break;
	    };
	    if(nwads<MAX_WADS) 
	      wadf[nwads++]=argv[i];
	 };
	 break;
       /* l = log */
       case('l'):
	 while(++i<argc) {
	    if(argv[i][0]=='-')  {
	       i--;
	       break;
	    };
	    log_file(argv[i],LOG_ALL,NULL);
	 };
	 break;
       /* r = record */
       case('r'):
	 if(++i>=argc) usage();
	 frec=fopen(argv[i],"wb");
	 break;
       /* p = record */
       case('p'):
	 if(++i>=argc) usage();
	 fplay=fopen(argv[i],"rb");
	 if(fplay) setvbuf(fplay,NULL,_IOFBF,4096);
	 break;
       /* m = map */
       case('m'):
	 if(++i>=argc) usage();
	 mapname=argv[i];
	 break;
       /* d=dimensions */
       case('d'):
	 if(i+2>=argc) usage();
	 width=atoi(argv[++i]);
	 height=atoi(argv[++i]);
	 break;
       /* b=bpp */
       case('b'):
	 if(++i>=argc) usage();
	 bpp=atoi(argv[i]);
	 break;
       /* hardness */
       case('h'):
	 if(++i>=argc) usage();
	 difficulty=atoi(argv[i]);
	 break;
       /* v=viewport */
       case('v'):
	 if(++i>=argc) usage();
	 vpdiv=atoi(argv[i]);
	 break;
       /* t=test */
      case('t'):
	 if(++i>=argc) usage();
	 testarg=argv[i];
	 break;
       /* ? = help */
       case('?'):
       default:
	 usage();
      };
   };

   if(!quiet) {
      /*setlinebuf(stdout);*/ /* if stdout is a socket, we'll want this */
      log_stdout();
   };
   logprintf(LOG_BANNER,'M',BANNER);
   
   if(nwads>0) {
      i=0;
      init_iwad(wadf[i++]);
      while(i<nwads) init_pwad(wadf[i++]);
   }
   else if(*mapname=='E'||*mapname=='e') {
     init_iwad("doom.wad");
     init_pwad("doom4dum.wad");
   }
   else {
     init_iwad("doom2.wad");
     init_pwad("doom4dum.wad");
   };
   init_wadhashing();

   init_sound(11025);
   init_textures();
   load_level(ld,mapname,difficulty,0);
   init_video(&width,&height,&bpp);
   video_winstuff(mapname,width,height);
   switch(bpp) {
#ifdef WANT_1BPP
   case(1):
      init_renderer=init_renderer8;
      render=render8;
      break;
#endif
#ifdef WANT_2BPP
   case(2):
      init_renderer=init_renderer16;
      render=render16;
      break;
#endif
#ifdef WANT_4BPP
   case(4):
      init_renderer=init_renderer32;
      render=render32;
      break;
#endif
   default:
      logfatal('D',"Unsupported BPP=%d",bpp);
   };
   init_input();
   set_playpal(0,video_setpal);

   if(no_objs) for(i=0;i<ldnthings(ld);i++) 
     if(i!=ld->player[0]) ldthingd(ld)[i].proto=NULL;
   
   td=ldthingd(ld)+(follow=ld->player[0]);

   init_view(&view);
   view.angle=0;
   view.x=td->x;
   view.y=td->y;
   if(td->sector==-1) 
     logprintf(LOG_ERROR,'D',"Can't find sector for player");
   view.sector=td->sector;
   sd=ldsectord(ld)+td->sector;
   if(vpdiv>1)
     init_renderer(width-width/vpdiv,height-height/vpdiv,width,height);
   else
     init_renderer(width,height,width,height);
   init_draw(width,height,bpp);

   init_gettables();
   set_gettable_special(1,&td->hits);

   banner=init_banner(64,0,width,10);
   wbanner=init_banner(height-12,0,width,3);
   bfont=init_font("STCFN%03d",128);
   if(have_lump("DUMBLOGO")) 
      add_to_banner(banner,get_misc_texture("DUMBLOGO"),width,64);
   if(have_lump("WAFFLE")) {
      LumpNum w=getlump("WAFFLE");
      add_to_banner(wbanner,NULL,width,0);
      add_text_to_banner(wbanner,bfont,
			 (const char *)load_lump(w),
			 get_lump_len(w));
   };

   if(testarg) {
      tdraw=get_misc_texture(testarg);
      logprintf(LOG_DEBUG,'D',"texture %s: real=%dx%d log=%dx%d",
		tdraw->name,tdraw->width,tdraw->height,
		1<<tdraw->log2width,1<<tdraw->log2height);
   };

   /* fill all video pages with our startup screen */
   /*if(have_lump("TITLEPIC")&&have_lump("DUMBLOGO")) {
      int i;
      Texture *t1=get_misc_texture("TITLEPIC");
      Texture *t2=get_misc_texture("DUMBLOGO");
      for(i=0;i<4;i++) {
	 fb=video_newframe();
	 draw_center(fb,t1);
	 draw_center(fb,t2);
	 video_updateframe(fb);
      };
   };*/
   
   tt=time(NULL);
   while(1) {
      PlayerInput in;
      if(ldthingd(ld)[follow].proto==NULL) follow=ld->player[0];
      thing_to_view(ld,follow,&view);
      render(fb=video_newframe(),ld,&view);
      draw_gettables(fb,width,height);
      update_banners(fb,1);
      if(tdraw) draw(fb,tdraw,0,0);
      video_updateframe(fb);
      poll_sound();
      if(fplay) {
	 if(feof(fplay)) {
	    logprintf(LOG_INFO,'D',"Demo ran out...");
	    fclose(fplay);
	    fplay=NULL;
	 }
	 else fread(&in,sizeof(in),1,fplay);
      };
      if(fplay==NULL) get_input(&in);
      if(frec) {
	 fwrite(&in,sizeof(in),1,frec);
	 fflush(frec);
      };
      if(in.quit) break;
      if(in.sideways) 
      	thingd_apply_sideways(td,INT_TO_FIXED(in.sideways)/(2*UNIT_SPEED));
      if(in.forward) 
      	thingd_apply_forward(td,INT_TO_FIXED(in.forward)/(2*UNIT_SPEED));
      if(in.rotate) 
      	thingd_apply_torque(td,INT_TO_FIXED(in.rotate)/(16*UNIT_SPEED));
      if(in.jump) 
      	thingd_apply_up(td,INT_TO_FIXED(in.jump)/UNIT_SPEED);
      if(in.shoot&&shframes<=0) {
	 /*fixed angle,elev;*/
	 shframes=4;
	 /*thing_autotarget(ld,ld->player[0],FIXED_PI/6);
	 logprintf(LOG_DEBUG,'D',"autotarget=%d",td->target);
	 thing_autoaim(ld,ld->player[0],FIXED_PI/6,&angle,&elev);
	 thing_shoot(ld,ld->player[0],9001,angle,elev,FIXED_PI/24,6);*/
	 use_selection(0,ld,ld->player[0]);
      };
      /* check for special functions */
      if(shframes<=0) {
	 if(in.select[0]) {
	    char buf[10];
	    shframes=4;
	    get_next_level_name(buf,ld->name);
	    logprintf(LOG_INFO,'D',"Jumping to next level (%s)",buf);
	    free_level(ld);
	    load_level(ld,buf,difficulty,0);
	    td=ldthingd(ld)+(follow=ld->player[0]);
	    set_gettable_special(1,&td->hits);
	    continue;
	 };
	 if(in.select[9]) {
	    int x,y;
	    unsigned int i;
	    const unsigned char *f=(const unsigned char *)fb;
	    FILE *fout;
	    logprintf(LOG_INFO,'D',"saving snapshot...");
	    fout=fopen("snapshot.ppm","w");
	    fprintf(fout,"P6\n%d %d\n%d\n",width,height,bpp==2?63:255);
	    for(y=0;y<height;y++) for(x=0;x<width;x++) 
	       switch(bpp) {
	       case(4):
		  putc(f[0],fout);
		  putc(f[1],fout);
		  putc(f[2],fout);
		  f+=4;
		  break;
	       case(2):
		  i=f[0]+((unsigned int)(f[1])<<8);
		  putc(((i>>10)&0x3e)|1,fout);
		  putc((i>>5)&0x3f,fout);
		  putc(((i&0x1f)<<1)|1,fout);
		  f+=2;
		  break;
	       case(1):
		  putc(0,fout);
		  putc(0,fout);
		  putc(0,fout);
		  f++;
		  break;
	       };
	    fclose(fout);
	    logprintf(LOG_INFO,'D',"snapshot done");
	 };
	 if(in.select[1]) {
	    shframes=4;
	    thing_hurl(ld,ld->player[0],8001,0,td->elev);
	 };
	 if(in.select[2]||in.w_sel) {
	    shframes=4;
	    /*thing_hurl(ld,ld->player[0],8002,0,td->elev);*/
	    rotate_selection(0);
	 };
	 if(in.select[3]) {
	    shframes=4;
	    thing_hurl(ld,ld->player[0],8004,0,td->elev);
	 };
	 if(in.select[4]) {
	    shframes=4;
	    thing_hurl(ld,ld->player[0],8005,0,td->elev);
	 };
	 if(in.select[5]) {
	    shframes=4;
	    thing_hurl(ld,ld->player[0],8006,0,td->elev);
	 };
	 if(in.select[6]) {
	    SoundBalance b=FLAT_BALANCE;
	    shframes=4;
	    if(!LUMPNUM_OK(sound_ln)) {
	       sound_ln=getlump("DSPLPAIN");
	       sound=(const unsigned char *)load_lump(sound_ln);
	    };
	    play_sound(sound+8,get_lump_len(sound_ln)-8,
		       b,11025);
	 };
      };
      /* back to ordinary main loop */
      player_apply_lookup(ld,ld->player[0],in.lookup/UNIT_SPEED);
      update_things(ld);
      update_map(ld);
      update_gettables(1);
      if(td->proto==NULL||td->hits<=0) {
	 logprintf(LOG_INFO,'D',"Hmmm... you seem to have died.");
	 free_level(ld);
	 load_level(ld,mapname,difficulty,0);
	 td=ldthingd(ld)+(follow=ld->player[0]);
	 set_gettable_special(1,&td->hits);
      };
      if(shframes>0) shframes--;
      frames++;
   };
   tt=time(NULL)-tt;
   logprintf(LOG_INFO,'D',"%d frames, %d seconds, %f fps",
	     frames,tt,(double)frames/tt);
   
   /*
   tt=time(NULL);
   for(view.height=sd->floor+(8<<12);view.height<sd->ceiling-(8>>12);view.height+=1<<11) {
      view.angle+=1<<9;
      render(fb=video_newframe(),ld,&view);
      video_updateframe(fb);
      frames++;
   };
   tt=time(NULL)-tt;
   logprintf(LOG_INFO,'D',"%ld frames, %ld seconds, %f fps",frames,tt,(double)frames/tt);
   */
   
   /*sleep(2);*/
   reset_input();
   reset_video();
   reset_sound();
   reset_gettables();
   free_level(ld);
   reset_textures();
   reset_wad();
   if(fplay) fclose(fplay);
   if(frec) fclose(frec);
   log_exit();
   return 0;
};










