#include <stdarg.h>
#include <stdio.h>

#include "lib/log.h"
#include "levdyn.h"
#include "things.h"
#include "updmap.h"

static int find_taggedsect(LevData *ld,int tag,int i) {
   while(i<ldnsectors(ld)) {
      if(ldsector(ld)[i].tag==tag) return i;
      i++;
   };
   return -1;
};

static int find_telespot(LevData *ld,int sector) {
   int i;
   const ThingDyn *td=ldthingd(ld);
   if(sector<0) return -1;
   for(i=0;i<ldnthings(ld);i++,td++)
      if(td->proto&&(td->proto->flags&PT_TELESPOT)&&td->sector==sector) 
	 return i;
   return -1;
};

/* th is where to make the fog, who is who is being teleported */
static void spawn_fog(LevData *ld,int th,int who,int prid) {
   const ThingDyn *td=ldthingd(ld)+th;
   const ThingDyn *tdwho=ldthingd(ld)+who;
   fixed x=td->x,y=td->y,z=td->z;
   x+=(9*fixcos(td->angle))/8;
   y+=(9*fixsin(td->angle))/8;
   z+=(3*tdwho->proto->height)/4;
   new_thing(ld,prid,x,y,z);
};


void thing_hit_wall(LevData *ld,int thing,int wall,int isback,WallHitType t)  {
   short front,back;
   front=ldline(ld)[wall].side[0];
   back=ldline(ld)[wall].side[1];
   if(front>=0) front=ldside(ld)[front].sector;
   if(back>=0) back=ldside(ld)[back].sector;

   /*logprintf(LOG_DEBUG,'O',"Thing %d hit wall %d back=%d t=%d",thing,wall,isback,t);*/
   
   /* this is just a kludge: it would be much better to go table-driven */
   switch(ldline(ld)[wall].type) {
      /* teleport */
   case(39):
   case(97):
      if(isback||t!=Crossed) break;
      if(!(ldthingd(ld)[thing].proto->flags&PT_PLAYER)) break;
      if(!find_event(ld,ML_THING,ME_TELEPORT,thing)) {
	 MapEvent *ev=insert_event(ld,ML_THING,ME_TELEPORT,thing);
	 int targsect=find_taggedsect(ld,ldline(ld)[wall].tag,0);
	 int targ=find_telespot(ld,targsect);
	 logprintf(LOG_DEBUG,'M',"teleport: sect=%d targ=%d",targsect,targ);
	 if(targ<0) targ=thing;
	 ev->start+=5;
	 ev->x=ldthingd(ld)[targ].x;
	 ev->y=ldthingd(ld)[targ].y;
	 ev->z=ldthingd(ld)[targ].z;
	 ev->angle=ldthingd(ld)[targ].angle;
	 ldthingd(ld)[thing].dx=ldthingd(ld)[thing].dy=0;
	 spawn_fog(ld,thing,thing,9014);
	 if(targ!=thing) spawn_fog(ld,targ,thing,9014);
      };
      break;
      /* remote doors */
   case(4):
   case(29):
   case(90):
   case(63):
   case(2):
   case(103):
   case(86):
   case(61):
   case(108):
   case(111):
   case(105):
   case(114):
   case(109):
   case(112):
   case(106):
   case(115):
   case(133):
   case(99):
   case(135):
   case(134):
   case(137):
   case(136):
      back=find_taggedsect(ld,ldline(ld)[wall].tag,0);
      /* fall through to manual door */

      /* manual doors */
   case(1):
   case(26):
   case(28):
   case(27):
   case(31):
   case(32):
   case(33):
   case(34):
   case(47):
   case(117):
   case(118):
      if(back<0) logprintf(LOG_ERROR,'M',
			   "Manual door with no back? (%d)",wall);
      else {
	 if(!find_active_event(ld,ML_SECTOR,ME_OPENDOOR,back)) {
	    unqueue_event(ld,ML_SECTOR,ME_OPENDOOR,back);
	    insert_event(ld,ML_SECTOR,ME_OPENDOOR,back);
	 };
	 unqueue_event(ld,ML_SECTOR,ME_CLOSEDOOR,back);
	 insert_event(ld,ML_SECTOR,ME_CLOSEDOOR,back)->start+=128;
      };
   };
};












