#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "lib/log.h"
#include "things.h"
#include "lib/fixed.h"

void player_apply_lookup(LevData *ld,int th,int lookup) {
   ThingDyn *td=ldthingd(ld)+th;
   td->elev=(td->elev+td->elev)/3;
   td->elev+=INT_TO_FIXED(lookup)/8;
};

static int appropriate_target(LevData *ld,int targ,int th,fixed arc)  {
   const ThingDyn *td=ldthingd(ld)+th;
   const ThingDyn *ttd=ldthingd(ld)+targ;
   fixed t;
   if(targ<0||targ==th) return 0;
   if(ttd->proto==NULL) return 0;
   if(!(ttd->proto->flags&PT_TARGET)) return 0;
   if(ttd->hits<=0) return 0; 
   t=fix_vec2angle(ttd->x-td->x,ttd->y-td->y)-td->angle;
   NORMALIZE_ANGLE(t);
   return t<arc/2||t>(FIXED_2PI-arc/2);
};

#define AUTOTARG_MAXDIST (512<<12)

void thing_autotarget(LevData *ld,int th,fixed arc) {
   ThingDyn *td=ldthingd(ld)+th;
   /* find a target */
   fixed minlen=AUTOTARG_MAXDIST;
   int i,is;
   td->target=-1;
   i=is=((unsigned)rand())%ldnthings(ld);
   do {
      if(appropriate_target(ld,i,th,arc)) {
	 fixed l=fix_pyth3d(ldthingd(ld)[i].x-td->x,
			    ldthingd(ld)[i].y-td->y,
			    ldthingd(ld)[i].z-td->z);
	 if(l<minlen)  {
	    minlen=l;
	    td->target=i;
	    };
      };
      i++;
      if(i>=ldnthings(ld)) i=0;
   } while(i!=is);
};

#define HEATSEEK_ARC (FIXED_PI/3) /* 60 degrees */

static void do_heatseek(LevData *ld,int th)  {
   ThingDyn *td=ldthingd(ld)+th;
   /* find a target */
   if(!appropriate_target(ld,td->target,th,HEATSEEK_ARC)) 
      thing_autotarget(ld,th,HEATSEEK_ARC);
   /* if we found one, try to steer towards it */
   if(td->target>=0)  {
      const ThingDyn *ttd=ldthingd(ld)+td->target;
      fixed speed=fix_pythagoras(td->dx,td->dy);
      fixed t=fix_vec2angle(ttd->x-td->x,ttd->y-td->y)-td->angle;
      fixed dz=ttd->z+ttd->proto->height/2-td->z;
      //logprintf(LOG_DEBUG,'O',"Obj %d is heatseeking %d",th,td->target);
      NORMALIZE_ANGLE(t);
      if(t<HEATSEEK_ARC/2) td->angle+=t/3;
      if(t>(FIXED_2PI-HEATSEEK_ARC/2)) td->angle+=(t-FIXED_2PI)/3;
      td->dx=td->dy=0;
      thingd_apply_forward(td,speed);
      if(dz>speed/3) td->dz=speed/3;
      else if(dz<-speed/3) td->dz=-speed/3;
      else td->dz=dz/3;
      td->elev=fix_vec2angle(speed,td->dz);
      /*logprintf(LOG_DEBUG,'O',"heatseek: dz=%f speed=%f elev=%f",
		  FIXED_TO_FLOAT(td->dz),FIXED_TO_FLOAT(speed),
		  FIXED_TO_FLOAT(td->elev));*/
   };
};

#define CAN_SEE_ARC (FIXED_PI/2) /* 90 degrees (should this be less?) */ 
int thing_can_see(LevData *ld,int th,int targ) {
   if(reject_thing_thing(ld,th,targ)) return 0;
   else {
      fixed a=fix_vec2angle(ldthingd(ld)[targ].x-ldthingd(ld)[th].x,
			    ldthingd(ld)[targ].y-ldthingd(ld)[th].y);
      a-=ldthingd(ld)[th].angle;
      NORMALIZE_ANGLE(a);
      if(a<CAN_SEE_ARC/2||a>FIXED_2PI-CAN_SEE_ARC/2) {
	 if(thing_can_shoot_at(ld,th,targ)) return 1;
	 else return 0;
      }
      else return 0;
   };
};


extern inline double pyth_sq(fixed x,fixed y) {
   double dx=FIXED_TO_FLOAT(x),dy=FIXED_TO_FLOAT(y);
   return dx*dx+dy*dy;
}; 
#define SQ(x) ((x)*(x))

/* for now, everyone hates players! */
#define emnity(p1,p2) ((p2)->flags&PT_PLAYER)

void thing_find_enemy(LevData *ld,int th) {
   ThingDyn *td=ldthingd(ld)+th;
   int targ,ts;
   const ThingDyn *ttd;
   /* random start ensures that monsters won't all fixate on one player */
   /* Should we check every object to see if it could be a target? */
   if(td->proto->flags&PT_USE_EMNITY) {
      targ=ts=((unsigned)rand())%ldnthings(ld);
      do {
	 ttd=ldthingd(ld)+targ;
	 if(ttd->proto&&
	    ttd->hits>0&&
	    emnity(td->proto,ttd->proto)&&
	    thing_can_see(ld,th,targ))
	    {
	       td->target=targ;
	       break;
	    };
	 targ++;
	 if(targ>=ldnthings(ld)) targ=0;
      } while (targ!=ts);
   }
   /* just check for live players */
   else {
      targ=ts=((unsigned)rand())%MAXPLAYERS;
      do {
	 ttd=ldthingd(ld)+ld->player[targ];
	 if(ld->player[targ]>=0&&
	    ttd->proto&&
	    ttd->hits>0&&
	    thing_can_see(ld,th,ld->player[targ])) {
	    td->target=ld->player[targ];
	    break;
	 };
	 targ++;
	 if(targ==MAXPLAYERS) targ=0;
      } while(targ!=ts) ;
   };
};

#define CLOSE_ENOUGH_ARC (FIXED_PI/16)
static void do_strategy(LevData *ld,int th) {
   ThingDyn *td=ldthingd(ld)+th;

   /* trash dead targets */ 
   if(td->target>=0&&ldthingd(ld)[td->target].proto==NULL) td->target=-1;

   /* need a new one? */
   if(td->target<0) {
      thing_find_enemy(ld,th);
      if(td->target>=0) 
	 logprintf(LOG_DEBUG,'O',"thing %d is out to get %d!",th,td->target);
   };
   
   /* chase after and attack our target */
   if(td->target>=0) {
      ThingDyn *ttd=ldthingd(ld)+td->target;
      fixed t=fix_vec2angle(ttd->x-td->x,ttd->y-td->y)-td->angle;
      double d2=pyth_sq(ttd->x-td->x,ttd->y-td->y);
      fixed attack_dist=td->proto->radius+ttd->proto->radius;
      double attd2;
      fixed dz=ttd->z-td->z;
      //logprintf(LOG_DEBUG,'O',"silly strategy for %d",th);
      NORMALIZE_ANGLE(t);
      if(td->proto->signals[TS_FIGHT]<0) attack_dist+=INT_TO_FIXED(4);
      else attack_dist+=FIXED_ONE;
      attd2=SQ(FIXED_TO_FLOAT(attack_dist));
      /* try fighting or shooting */
      if(t<CLOSE_ENOUGH_ARC/2||t>FIXED_2PI-CLOSE_ENOUGH_ARC/2) {
	 if(reject_thing_thing(ld,th,td->target));
	 else if(td->proto->signals[TS_FIGHT]>=0&&d2<=attd2)
	    thing_send_sig(ld,th,TS_FIGHT);
	 else if(!thing_can_shoot_at(ld,th,td->target));
	 else if(td->proto->flags&PT_FAST_SHOOTER) 
	    thing_send_sig(ld,th,TS_SHOOT);
	 else if(!(rand()&7)) 
	    thing_send_sig(ld,th,TS_SHOOT);
      }
      else if(t<FIXED_PI) thingd_apply_torque(td,td->proto->realmass/10);
      else thingd_apply_torque(td,-td->proto->realmass/10);
      /* move towards foe */
      if(d2>attd2)
	 thingd_apply_forward(td,td->proto->realmass/2);
      if(td->proto->flags&PT_CAN_FLY) {
	 if(dz>0)
	    thingd_apply_up(td,td->proto->realmass/2);
	 if(dz<-td->proto->height);
	    thingd_apply_up(td,-td->proto->realmass/2);
      };
      /* give up if they seem dead */
      if(!(ttd->proto->flags&PT_BEASTIE)) td->target=-1;
   };
};

static void spawn_or_hurl(LevData *ld,int th,int prid)  {
   ThingDyn *td=ldthingd(ld)+th;
   new_thing(ld,prid,td->x,td->y,td->z);
};

/* wake anyone who can see me */
void thing_wake_others(LevData *ld,int th) {
   int i;
   for(i=0;i<ldnthings(ld);i++) {
      const ThingDyn *td=ldthingd(ld)+i;
      const ThingPhase *tph=td->phase_tbl+td->phase;
      if(i!=th&&td->proto&&(tph->flags&TPH_IDLE)&&thing_can_see(ld,i,th))
	 thing_send_sig(ld,i,TS_DETECT);
   };
};

/* take damage, potentially ouching, dying, or exploding */
void thing_take_damage(LevData *ld,int th,int dmg) {
   ThingDyn *td=ldthingd(ld)+th;
   /* check if we explode before checking hits */
   if(td->proto->flags&PT_EXPLOSIVE) {
      td->dx=td->dy=td->dz=0; /* stop explosions from being knocked around */
      thing_send_sig(ld,th,TS_EXPLODE);
   };
   /* if hits<0 either we're already dead, 
    * or we were invulernable to begin with (proto->hits==-1)
    */
   if(td->hits<0) return;
   /* TODO: check temporary invulnerability */
   
   logprintf(LOG_DEBUG,'O',"thing %d took %d damage (%d/%d hits)",
	     th,dmg,td->hits,(int)(td->proto->hits));
   td->hits-=dmg;
   /* we survived */
   if(td->hits>0) {
      thing_send_sig(ld,th,TS_OUCH);
      return;
   };
   /* we didn't: how sad, never mind */
   if(td->hits<-td->proto->hits&&thing_send_sig(ld,th,TS_EXPLODE)!=-1) 
      return; 
   thing_send_sig(ld,th,TS_DIE);
};

/* apply torque so that td turns towards angle 
   intensity should be between 1 and 0 
   this fn is not currently used: I don't know whether or not it works
*/
/*void thingd_apply_torque_towards(ThingDyn *td,fixed angle,fixed intensity) {
   angle-=td->angle;
   NORMALIZE_ANGLE(angle);
   if(angle<FIXED_PI) td->dangle+=fixmul(angle,intensity);
   else td->dangle+=fixmul(FIXED_2PI-angle,intensity);
};*/

static int unowned_owner(LevData *ld,int th) {
   int o;
   while((o=ldthingd(ld)[th].owner)>=0) th=o;
   return th;
};

/* do (explosion-like) damage to nearby monsters */
#define EXPLODE_RADIUS (FIXED_ONE*8)
static void do_damage(LevData *ld,int th,fixed arc) {
   int targ,myowner=unowned_owner(ld,th);
   const ThingDyn *td=ldthingd(ld)+th;
   ThingDyn *ttd=ldthingd(ld);
   arc/=2;
   for(targ=0;targ<ldnthings(ld);targ++,ttd++) {
      fixed angle,dist,dmul;
      int dmg;
      if(ttd->proto==NULL||ttd->proto->hits==-1||targ==th) continue;
      angle=fix_vec2angle(ttd->x-td->x,ttd->y-td->y);
      NORMALIZE_ANGLE(angle);
      /* this thing completely behind effect, anyway? */
      if(angle>arc&&angle<(FIXED_2PI-arc)) continue;
      /* no, check distance (lots of guesstimation going on here) */
      dist=fix_pyth3d(ttd->x-td->x,ttd->y-td->y,ttd->z-td->z);
      dist-=(ttd->proto->radius+ttd->proto->height)/2;
      /* damage calculation */
      if(dist>EXPLODE_RADIUS) continue;
      dmul=fixdiv(EXPLODE_RADIUS-dist,EXPLODE_RADIUS);
      dmul=fixmul(dmul,dmul);
      dmg=FIXED_TO_INT(fixmul(INT_TO_FIXED(td->proto->damage),dmul));
      if(dmg<=0) continue;
      /* OK, we've hurt them: what happens? */
      /* 1) they get pissed off with us (unless we are they) */
      if(myowner!=targ) ttd->target=myowner; 
      /* 2) knockback */
      thingd_apply_polar(ttd,INT_TO_FIXED(dmg)/16,angle);
      if(td->z<=ttd->z)
	 thingd_apply_up(ttd,INT_TO_FIXED(dmg)/16);
      else if(td->z<ttd->z+ttd->proto->height)
	 thingd_apply_up(ttd,INT_TO_FIXED(dmg)/32);
      /* 3) they take damage */
      thing_take_damage(ld,targ,dmg);
   };
};

/* do damage just to target */
static void do_melee_damage(LevData *ld,int th,fixed knockback) {
   ThingDyn *td=ldthingd(ld)+th;
   ThingDyn *ttd=ldthingd(ld)+td->target;
   fixed dist;
   int myowner;
   if(td->target<0||ttd->proto==NULL) return;
   dist=fix_pyth3d(td->x-ttd->x,td->y-ttd->y,td->z-ttd->z);
   if(dist>td->proto->radius+td->proto->height+
      ttd->proto->radius+ttd->proto->height) {
      logprintf(LOG_DEBUG,'O',"melee: %d's attack on %d failed",th,td->target);
      return;
   };
   /* piss them off */
   myowner=unowned_owner(ld,th);
   if(myowner!=td->target) ttd->target=myowner;
   /* knock them back */
   thingd_apply_polar(ttd,td->proto->damage*knockback,td->angle);
   /* actually do the damage */
   thing_take_damage(ld,td->target,td->proto->damage);
};

void thing_autoaim(LevData *ld,int th,fixed arc,fixed *angle,fixed *elev) {
   ThingDyn *td=ldthingd(ld)+th;
   *angle=*elev=FIXED_ZERO;
   /* figure out angle to target */
   if(td->target>=0) {
      const ThingDyn *ttd=ldthingd(ld)+td->target;
      fixed t=fix_vec2angle(ttd->x-td->x,ttd->y-td->y)-td->angle;
      fixed d=fix_pythagoras(ttd->x-td->x,ttd->y-td->y);
      fixed e=fix_vec2angle(d,(ttd->z+SHOOT_HEIGHT(ttd)) -
			    (td->z+SHOOT_HEIGHT(td)));
      NORMALIZE_ANGLE(t);
      NORMALIZE_ANGLE(e);
      if(t<arc/2) {*angle=t; *elev=e;};
      if(t>(FIXED_2PI-arc/2)) {*angle=t; *elev=e;};
   };
};

/*
  TPH_SHOOT encompasses many activities: it is supposed to do
  whatever shooting type activity is appropriate to the thing, or if none
  to just spawn a spawn1 object
*/
#define AUTOAIM_ARC (FIXED_PI/4)
static void do_shoot(LevData *ld,int th) {
   fixed angle=FIXED_ZERO,elev=FIXED_ZERO;
   ThingDyn *td=ldthingd(ld)+th;
   /* figure out angle to target */
   thing_autoaim(ld,th,AUTOAIM_ARC,&angle,&elev);
   /* do it */
   if(td->proto->flags&PT_SHOOTER) 
      thing_shoot(ld,th,td->proto->spawn1,angle,elev,
		  td->proto->shootarc,td->proto->shootnum);
   else
      thing_hurl(ld,th,td->proto->spawn1,angle,elev);
};

void thing_enter_phase(LevData *ld,int th,int ph)  {
   ThingDyn *td=ldthingd(ld)+th;
   const ThingPhase *nu=td->phase_tbl+ph;
   const ThingPhase *old=td->phase_tbl+td->phase;
   /* do post-phase effects */
   if(old->flags&TPH_DESTROY)  {
      td->proto=NULL;
      return;
   };
   if(old->flags&TPH_BECOME) {
      logprintf(LOG_DEBUG,'O',"Thing %d becomes type %d; old type was %d",
		th,(int)(td->proto->become1),(int)(td->proto->id));
      td->proto=find_protothing(td->proto->become1);
      if(td->proto==NULL) return;
      td->phase_tbl=find_first_thingphase(td->proto->phase_id);
      td->hits=td->proto->hits;
      ph=td->proto->signals[TS_INIT];
      nu=td->phase_tbl+ph;
   };
   if(old->flags&TPH_BECOME2) {
      logprintf(LOG_DEBUG,'O',"Thing %d becomes type %d; old type was %d",
		th,(int)(td->proto->become2),(int)(td->proto->id));
      td->proto=find_protothing(td->proto->become2);
      if(td->proto==NULL) return;
      td->phase_tbl=find_first_thingphase(td->proto->phase_id);
      td->hits=td->proto->hits;
      ph=td->proto->signals[TS_INIT];
      nu=td->phase_tbl+ph;
   };
   /* update phase */
   td->phase=ph;
   td->phase_wait=nu->wait;
   /* do pre-phase effects */
   if(nu->flags&TPH_EXPLODE) do_damage(ld,th,FIXED_2PI);
   if(nu->flags&TPH_MELEE) 
      do_melee_damage(ld,th,
		      td->proto->flags&PT_BEASTIE?FIXED_ZERO:FIXED_ONE/2);
   /*if(nu->flags&TPH_BDMG) do_melee_damage(ld,th,FIXED_ONE/2);*/
   if(nu->flags&TPH_HEATSEEK) do_heatseek(ld,th);
   if(nu->flags&TPH_STRATEGY) do_strategy(ld,th);
   if(nu->flags&TPH_NOISY) thing_wake_others(ld,th);
   if(nu->flags&TPH_SHOOT) do_shoot(ld,th);
   if(nu->flags&TPH_SPAWN2||((nu->flags&TPH_RSPAWN2)&&(rand()&1)))
     spawn_or_hurl(ld,th,td->proto->spawn2);
};

int thing_send_sig(LevData *ld,int th,ThingSignal ts)  {
   ThingDyn *td=ldthingd(ld)+th;
   /* if 
       1) we respond to this type of signal, and
       2) we aren't already pending a higher priority signal, and
       3) signals aren't locked out for this phase,
      then
       make ts pending
      */
   if(td->proto->signals[ts]>=0&& 
      td->pending_signal<ts&& 
      !(td->phase_tbl[td->phase].flags&TPH_NOSIGS))
      td->pending_signal=ts;
   else return -1;
   return 0;
};

void thingd_apply_polar(ThingDyn *td,fixed f,fixed a) {
   if(td->proto->flags&PT_INF_MASS) return;
   td->dx+=fixmul(fixcos(a),fixdiv(f,td->proto->realmass));
   td->dy+=fixmul(fixsin(a),fixdiv(f,td->proto->realmass));
};
void thingd_apply_torque(ThingDyn *td,fixed t) {
   if(td->proto->flags&PT_INF_MASS) return;
   td->dangle+=fixdiv(t,td->proto->realmass);
};
void thingd_apply_up(ThingDyn *td,fixed f) {
   if(td->proto->flags&PT_INF_MASS) return;
   td->dz+=fixdiv(f,td->proto->realmass);
};

void thing_to_view(LevData *ld,int th,View *v)  {
   ThingDyn *td=ldthingd(ld)+th;
   v->height=td->z+SHOOT_HEIGHT(td);
   v->angle=td->angle;
   NORMALIZE_ANGLE(v->angle);
   v->x=td->x;
   v->y=td->y;
   v->horizon=-fixsin(td->elev);
   v->sector=td->sector;
};

/* the angle between each rotation of a sprite in the wadfile */
#define THETA (FIXED_PI/4)
void thing_rotate_image(LevData *ld,int th,fixed angle)  {
   ThingDyn *td=ldthingd(ld)+th;
   char rot;
   angle-=td->angle+FIXED_PI;
   angle+=THETA/2;
   NORMALIZE_ANGLE(angle);
   rot='1'+angle/THETA;
   if(rot<'1'||rot>'8')
     rot='1';
   td->image=find_phase_sprite(td->proto,td->phase,rot);
   if(td->image&&td->image->name[7]==rot) td->mirror_image=1;
   else td->mirror_image=0;
};





