#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "lib/log.h"
#include "levdyn.h"
#include "animtex.h"
#include "updmap.h"
#include "things.h"

static fixed min_adjacent_ceiling(LevData *ld,short sector)  {
   int wall;
   fixed f=FIXED_MAX;
   for(wall=0;wall<ldnlines(ld);wall++)  {
      fixed g=FIXED_MAX;
      short front=ldline(ld)[wall].side[0];
      short back=ldline(ld)[wall].side[1];
      if(back<0) continue;
      front=ldside(ld)[front].sector;
      back=ldside(ld)[back].sector;
      if(front==sector) g=ldsectord(ld)[back].ceiling;
      else if(back==sector) g=ldsectord(ld)[front].ceiling;
      if(g<f) f=g;
   };
   return f;
};

static void init_sector_event(LevData *ld,MapEvent *me) {
   fixed speed=4<<12;
   SectorDyn *sd=ldsectord(ld)+me->entity;
   switch(me->type)  {
   case(ME_OPENDOOR):
      me->term=min_adjacent_ceiling(ld,me->entity)-(6<<12);
      me->delta=speed;
      break;
   case(ME_CLOSEDOOR):
      me->term=sd->floor;
      me->delta=-speed;
      break;
   default:
      logprintf(LOG_ERROR,'M',"init_sector_event: strange eventtype (%d)",
		me->type); 
   };
};

static void do_sector_event(LevData *ld,MapEvent *me) {
   SectorDyn *sd=ldsectord(ld)+me->entity;
   switch(me->type)  {
   case(ME_OPENDOOR):
   case(ME_CLOSEDOOR):
      if((me->delta<0&&sd->ceiling<=me->term)||
	 (me->delta>0&&sd->ceiling>=me->term)) {
	 /* terminate the move */
	 sd->ceiling=me->term;
	 me->type=ME_NONE;
      }
      else sd->ceiling+=me->delta;
      break;
   default:
      logprintf(LOG_ERROR,'M',"do_sector_event: strange eventtype (%d)",
		me->type); 
   };
};

static void init_thing_event(LevData *ld,MapEvent *me) {
   ThingDyn *td=ldthingd(ld)+me->entity;
   me->x=td->x;
   me->y=td->y;
   me->z=td->z;
   me->angle=td->angle;
};

static void do_thing_event(LevData *ld,MapEvent *me) {
   ThingDyn *td=ldthingd(ld)+me->entity;
   switch(me->type)  {
   case(ME_TELEPORT):
      logprintf(LOG_DEBUG,'M',"do_thing_event: teleporting %d",me->entity);
      td->x=me->x;
      td->y=me->y;
      td->z=me->z;
      td->angle=me->angle;
      thingd_findsector(ld,td);
      me->type=ME_NONE;
      break;
   default:
      logprintf(LOG_ERROR,'M',"do_thing_event: strange eventtype (%d)",
		me->type); 
   };
};

static MapEvent bogus_event;

MapEvent *insert_event(LevData *ld,
		       MapLumpType lumptype,
		       MapEventType etype,
		       int entity) {
   int i;
   MapEvent *me;
   for(i=0;i<MAXEVENTS;i++)
      if(!EVENTVALID(ld->event[i])) break;
   if(i>=MAXEVENTS) {
      logprintf(LOG_ERROR,'M',"No free event in insert_event(%d,%d,%d)",
		lumptype,etype,entity);
      return &bogus_event;
   };
   me=ld->event+i;
   /* init event */
   logprintf(LOG_DEBUG,'M',"insert_event(%d,%d,%d)",lumptype,etype,entity);
   memset(me,0,sizeof(MapEvent));
   me->ltype=lumptype;
   me->type=etype;
   me->entity=entity;
   me->start=ld->map_ticks+1;
   me->stop=me->start+65536;
   /* now event-specific initialisation */
   switch(me->ltype) {
   case(ML_SECTOR): init_sector_event(ld,me); break;
   case(ML_THING): init_thing_event(ld,me); break;
   default:
      logprintf(LOG_ERROR,'M',"Strange lumptype %d in insert_event",lumptype);
   };
   return me;
};

int unqueue_event(LevData *ld,
		  MapLumpType lumptype,
		  MapEventType etype,
		  int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) {
	 logprintf(LOG_DEBUG,'M',"event %d unqueued",i);
	 j++;
	 memset(ld->event+i,0,sizeof(MapEvent));
      };
   };
   return j;
};

int find_active_event(LevData *ld,
		      MapLumpType lumptype,
		      MapEventType etype,
		      int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].start>ld->map_ticks) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) j++;
   };
   return j;
};

int find_event(LevData *ld,
		      MapLumpType lumptype,
		      MapEventType etype,
		      int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) j++;
   };
   return j;
};


static void um_sides(LevData *ld) {
   int i;
   SideDyn *sd=ldsided(ld);   
   for(i=0;i<ldnsides(ld);i++,sd++)  {
      /* update textures (if any) */
      if(sd->uanim>=0) sd->utex=get_anim_texture(sd->uanim,ld->map_ticks);
      if(sd->manim>=0) sd->mtex=get_anim_texture(sd->manim,ld->map_ticks);
      if(sd->lanim>=0) sd->ltex=get_anim_texture(sd->lanim,ld->map_ticks);
   };
};

static void um_sectors(LevData *ld) {
   int i;
   SectorDyn *sd=ldsectord(ld);
   for(i=0;i<ldnsectors(ld);i++,sd++)  {
      /* update animated floors & ceilings (if any) */
      if(sd->fanim>=0) sd->ftex=get_anim_texture(sd->fanim,ld->map_ticks);
      if(sd->canim>=0) sd->ctex=get_anim_texture(sd->canim,ld->map_ticks);
   };
};

static void um_events(LevData *ld) {
   int i;
   MapEvent *me;
   for(i=0,me=ld->event;i<MAXEVENTS;i++,me++) {
      if(!EVENTVALID(*me)||me->start>ld->map_ticks) continue;
      /* has event passed its best before date? */
      if(me->stop<ld->map_ticks) {
	 logprintf(LOG_DEBUG,'M',"event %d passed its stop date",i);
	 memset(me,0,sizeof(MapEvent));
	 continue;
      };
      /* should event run now? */
      if(me->start<=ld->map_ticks) switch(me->ltype) {
      case(ML_SECTOR): do_sector_event(ld,me); break;
      case(ML_THING): do_thing_event(ld,me); break;
      default:
	 logprintf(LOG_ERROR,'M',"Strange lumptype %d in um_events",
		   me->ltype);
      };
   };
};

void update_map(LevData *ld) {
   ld->map_ticks++;
   um_events(ld);
   um_sides(ld);
   um_sectors(ld);
};







