#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <vga.h>

#include "lib/safem.h"
#include "lib/log.h"
#include "video.h"
#include <ggi/libggi.h>

/* MIN_PAGES should normally be 2, but set it to 1 to force direct video */

#define MAX_PAGES 3
#define MIN_PAGES 2

static int pagelen=0;
static char *pagev=NULL;

static int mywidth,myheight;

static ggi_visual_t vis;

void video_preinit(void) {

   ggiInit();
   vis=ggiOpen(NULL);
   ggiSetFocus(vis);

};

void init_video(int *_width,int *_height,int *_bpp) {

   int err;
   if(pagev!=NULL) safe_free(pagev);
   pagev=NULL;

   switch(*_bpp) {
    case(4):
      err=ggiSetGraphMode(vis,*_width,*_height,*_width,*_height,GT_32BIT);
      break;
    case(2):
      err=ggiSetGraphMode(vis,*_width,*_height,*_width,*_height,GT_16BIT);
      break;
    case(1):
      err=ggiSetGraphMode(vis,*_width,*_height,*_width,*_height,GT_8BIT);
      break;
    default:
      logfatal('V',"Bad BPP (%d) in init_video",*_bpp);
   };

   if (err) logfatal('V',"GGI:SetGraphMode failed.");

   mywidth =*_width =ggiGetInfo(vis)->fb.width;
   myheight=*_height=ggiGetInfo(vis)->fb.height;
   
   pagelen=(*_width)*(*_height)*(*_bpp);

   logprintf(LOG_INFO,'V',"init_video in %s mode pagelen=%d","copying",pagelen);
   pagev=safe_malloc(pagelen);
};

void reset_video(void) {
   if(pagev!=NULL) safe_free(pagev);
   pagev=NULL;
   ggiExit();
};

#define SHLPAL 8L
void video_setpal(unsigned char idx,
		  unsigned char red,
		  unsigned char green,
		  unsigned char blue) {
   ggi_color gcol;
   gcol.r=red   << SHLPAL;
   gcol.g=green << SHLPAL;
   gcol.b=blue  << SHLPAL;

   ggiSetPaletteVec(vis,idx,1,&gcol);

};

void *video_newframe(void) {
   return pagev;
};

void video_updateframe(void *v) {

	ggiPutBox(vis,0,0,mywidth,myheight,v);
	ggiFlush(vis);	/* render NOW */
};

void video_winstuff(const char *desc,int xdim,int ydim) {};

#include "input.h"

static PlayerInput inca;

void get_input(PlayerInput *in) {

   static int speed=UNIT_SPEED;
   int evmask;
   ggi_event ev;
   struct timeval t={0,0};
   
   evmask=emKeyPress|emKeyRelease|emPointer;

   memcpy(in,&inca,sizeof(PlayerInput)); 
   while (ggiEventPoll(vis,evmask,&t))
   {
	/*ggiEventPoll(vis,evmask,NULL);*/

	do {
		ggiEventRead(vis,&ev,evmask);
	} while(! ( (1 << ev.any.type) & evmask ) );

	switch(ev.any.type) { 
	    case evKeyPress:
		switch(KTYP(U(ev.key.sym)))
		{ case KT_LATIN:
		  case KT_LETTER:
		  	switch(KVAL(U(ev.key.sym)))
		  	{
			  case ' ': /* space */
		  		inca.jump=UNIT_SPEED;break;
			  case '\x1b': /* esc */
		 	 	inca.quit=1;break;
			  case ',':
			  	inca.sideways+=speed;break;
			  case '.':
		  		inca.sideways-=speed;break;
			  case '0':inca.select[0]=1;break;
			  case '1':inca.select[1]=1;break;
			  case '2':inca.select[2]=1;break;
			  case '3':inca.select[3]=1;break;
			  case '4':inca.select[4]=1;break;
			  case '5':inca.select[5]=1;break;
			  case '6':inca.select[6]=1;break;
			  case '7':inca.select[7]=1;break;
			  case '8':inca.select[8]=1;break;
			  case '9':inca.select[9]=1;break;
			  default: printf("latin sym=%4x code=%4x (%c)\n",ev.key.sym,ev.key.code,KVAL(U(ev.key.sym)));break;
		  	}
		  	break;
		  default:
		  switch(U(ev.key.sym))
		  {
		    case K_UP: /* CrSr up */
		    case K_P8: /* Keypad 8 */
			inca.forward+=speed;break;
		    case K_DOWN: /* CrSr down */
		    case K_P2: /* Keypad 2 */
	  		inca.forward-=speed;break;
		    case K_RIGHT: /* CrSr right */
		    case K_P6: /* CrSr right */
			inca.rotate-=speed;break;
		    case K_LEFT: /* CrSr left */
		    case K_P4: /* CrSr left */
	  		inca.rotate+=speed;break;
	
		    case K_ENTER: /* enter */
		  	inca.shoot=1;break;
		    case K_P9:
		    case K_P7:
	  		inca.lookup+=UNIT_SPEED;break;
		    case K_P3:
		    case K_P1:
	  		inca.lookup-=UNIT_SPEED;break;
		    case K_NORMAL_SHIFT:
		  	speed=2*UNIT_SPEED;break;

		    default: printf("sym=%4x code=%4x\n",ev.key.sym,ev.key.code);break;
		  }
		}
		break;
	    case evKeyRelease:
		switch(KTYP(U(ev.key.sym)))
		{ case KT_LATIN:
		  case KT_LETTER:
		  	switch(KVAL(U(ev.key.sym)))
		  	{
			  case ' ': /* space */
		  		inca.jump=0;break;
			  case '\x1b': /* esc */
		 	 	inca.quit=0;break;
			  case ',':
			  case '.':
		  		inca.sideways=0;break;
			  case '0':inca.select[0]=0;break;
			  case '1':inca.select[1]=0;break;
			  case '2':inca.select[2]=0;break;
			  case '3':inca.select[3]=0;break;
			  case '4':inca.select[4]=0;break;
			  case '5':inca.select[5]=0;break;
			  case '6':inca.select[6]=0;break;
			  case '7':inca.select[7]=0;break;
			  case '8':inca.select[8]=0;break;
			  case '9':inca.select[9]=0;break;
		  	}
		  	break;
		  default:
		  switch(U(ev.key.sym))
		  {
		    case K_UP: /* CrSr up */
		    case K_P8: /* Keypad 8 */
		    case K_DOWN: /* CrSr down */
		    case K_P2: /* Keypad 2 */
	  		inca.forward=0;break;
		    case K_RIGHT: /* CrSr right */
		    case K_P6: /* CrSr right */
		    case K_LEFT: /* CrSr left */
		    case K_P4: /* CrSr left */
	  		inca.rotate=0;break;
	
		    case K_ENTER: /* enter */
		  	inca.shoot=0;break;
		    case K_P9:
		    case K_P7:
		    case K_P3:
		    case K_P1:
	  		inca.lookup=0;break;
		    case K_NORMAL_SHIFT:
		  	speed=UNIT_SPEED;break;
		  }
		}
		break;
	   case evPtrButtonPress:
		if (ev.pbutton.button&1)
	    		inca.forward=UNIT_SPEED;
		if (ev.pbutton.button&6)
	    		inca.shoot=1;
	    	break;
	   case evPtrButtonRelease:
		if (ev.pbutton.button&1)
	    		inca.forward=0;
		if (ev.pbutton.button&6)
	    		inca.shoot=0;
	    	break;
	}
	/* Now for "only now" effects.
	 */
	memcpy(in,&inca,sizeof(PlayerInput)); 
	switch(ev.any.type) { 
	    case evPtrRelative:
	    	in->rotate =-ev.pmove.x;
	    	in->forward=-ev.pmove.y;
	    	break;
	}
   }
};

void init_input(void) { 
   memset(&inca,0,sizeof(PlayerInput)); 
};

void reset_input(void) { };


