#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "dumb/prothing.h"
#include "dumb/gettable.h"

#define NAMELEN 32
#define LOTS 8192

FILE *fin;

int line=0,tok_ungot=0;
char tokbuf[256]="";
char cppfile[256]="stdin";

typedef struct {
   ProtoThing pt;
   char name[NAMELEN];
} ProtoThingRec;

typedef struct {
   Gettable g;
   char name[NAMELEN];
} GettableRec;

typedef struct {
   ThingPhase *tp;
   char *tpname;
   char name[NAMELEN];
   int nphases;
   LE_int16 signals[NUM_THINGSIGS];
} ThingPhaseRec;

int nprotos=0,nphasetbls=0,ngetts=0;
ProtoThingRec *protos;
ThingPhaseRec *phases;
GettableRec *getts;

int top_assigned_proto_id=16385;

void chkuniqueid(int id) {
   int i;
   for(i=nprotos-2;i>=0;i--)
      if(protos[i].pt.id==id) {
	 printf("ptcomp: line %d (%s): ID %d not unique (clashes with %s)\n",
		line,cppfile,id,protos[i].name);
	 exit(2);
      };
};

#define unget_token() tok_ungot++;

const char *next_token(void) {
   char *s=tokbuf;
   int i;
   if(line==0) line=1;
   if(tok_ungot) {
      tok_ungot=0;
      return tokbuf;
   };
   memset(tokbuf,0,256);
   while(1) {
      if(feof(fin)) return NULL;
      i=getc(fin);
      *s = (char)i;
      if(*s!=' '&&*s!='\t'&&*s!='\r'&&i!=-1) break;
   };
   if(*s=='#') {
      while((*s++=getc(fin))!='\n') if(feof(fin)) return NULL;
      line=atoi(tokbuf+1);
      s=strchr(tokbuf,'"');
      if(s) {
	 char *t=cppfile;
	 s++;
	 while(*s!='"') *t++=*s++;
	 *t=0;
	 if(*cppfile==0) strcpy(cppfile,"stdin");
      };
      /*printf("line=%d file=(%s)\n",line,cppfile);*/
      s=tokbuf;
      *s=0;
      return "\n";
   };
   if(*s=='\n') {
      line++;
      *s=0;
      return "\n";
   };
   do {
      if(feof(fin)) return tokbuf;
      s++;
      i=getc(fin);
      *s=(char)i;
   } while(*s!=' '&&*s!='\t'&&*s!='\r'&&i!=-1&&*s!='\n'&&*s!='#');
   if(*s=='\n'||*s=='#') ungetc(*s,fin);
   *s=0;
   return tokbuf;
};

void synerr(const char *detail) {
   printf("ptcomp: syntax error near line %d (%s), token '%s'\n",
	  line,cppfile,tokbuf);
   if(detail) printf("        %s\n",detail);
   exit(3);
};
void err(const char *detail) {
   printf("ptcomp: fatal error near line %d (%s)\n",line,cppfile);
   if(detail) printf("        %s\n",detail);
   exit(3);
};

ThingPhaseRec *find_ph_tbl(const char *s) {
   int i;
   for(i=0;i<nphasetbls;i++) 
      if(!strcasecmp(phases[i].name,s))
	 return phases+i;
   return NULL;
};
ThingPhaseRec *parm_ph_tbl(void) {
   const char *s=next_token();
   ThingPhaseRec *tpr;
   if(s==NULL||*s=='\n') synerr("phase table name parameter expected");
   tpr=find_ph_tbl(s);
   if(!tpr) synerr("phase table name unrecognised");
   return tpr;
};
int parm_phase(ThingPhaseRec *p) {
   int i;
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("phase name parameter expected");
   for(i=0;i<p->nphases;i++) 
      if(!strcasecmp(p->tpname+i*NAMELEN,s))
	 return i;
   synerr("phase name unrecognised");
   return -1;
};
int parm_proto(void) {
   int i;
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("proto name parameter expected");
   if(isdigit(*s)) return atoi(s);
   for(i=0;i<nprotos;i++) 
      if(!strcasecmp(protos[i].name,s))
	 return protos[i].pt.id;
   synerr("proto name unrecognised");
   return -1;
};
int parm_gett(void) {
   int i;
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("gettable name parameter expected");
   if(isdigit(*s)) return atoi(s);
   for(i=0;i<ngetts;i++) 
      if(!strcasecmp(getts[i].name,s))
	 return i;
   synerr("gettable name unrecognised");
   return -1;
};

void parm_str(char *buf,int n) {
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("string parameter expected");
   strncpy(buf,s,n-1);
   buf[n-1]=0;
};
char parm_ch(void) {
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("character parameter expected");
   return *s;
};
int parm_num(void) {
   const char *s=next_token();
   if(s==NULL||*s=='\n') synerr("integer parameter expected");
   return atoi(s);
};

void parm_msg(char *buf,int n) {
   char *s;
   parm_str(buf,n);
   for(s=buf;*s;s++) if(*s=='_') *s=' ';
};

void gettcomp(void) {
   const char *s;
   GettableRec *p;
   /* make new gettable */
   p=getts+(ngetts++);
   if(ngetts>=LOTS) err("too many gettables");
   memset(p,0,sizeof(GettableRec));
   p->g.bulletkind=-1;
   p->g.usenum=1;
   /* one parameter, name */
   s=next_token();
   if(s==NULL||*s=='\n') synerr("name expected after Gettable");
   strncpy(p->name,s,NAMELEN-1);
   /* now the info */
   while(1) {
      s=next_token();
      if(s==NULL) return;
      else if(*s=='\n');
      else if(!strcasecmp(s,"Gun")) p->g.flags|=GK_SHOOT;
      else if(!strcasecmp(s,"Melee")) p->g.flags|=GK_MELEE;
      else if(!strcasecmp(s,"WepSel")) p->g.flags|=GK_WEPSELECT;
      else if(!strcasecmp(s,"SpeSel")) p->g.flags|=GK_SPESELECT;
      else if(!strcasecmp(s,"GotThe")) p->g.flags|=GK_GOT_THE;
      else if(!strcasecmp(s,"GotA")) p->g.flags|=GK_GOT_A;
      else if(!strcasecmp(s,"KeyType")) p->g.key=parm_num();
      else if(!strcasecmp(s,"Decay")) p->g.decay=parm_num();
      else if(!strcasecmp(s,"Special")) p->g.special=parm_num();
      else if(!strcasecmp(s,"Collect")) p->g.collect=parm_num();
      else if(!strcasecmp(s,"Initial")) p->g.initial=parm_num();
      else if(!strcasecmp(s,"Usenum")) p->g.usenum=parm_num();
      else if(!strcasecmp(s,"AddAmmo")) p->g.bulletadd=parm_num();
      else if(!strcasecmp(s,"Ammo")) p->g.bulletkind=parm_gett();
      else if(!strcasecmp(s,"Shoots")) p->g.bullettype=parm_proto();
      else if(!strcasecmp(s,"Icon")) parm_str(p->g.iconname,10);
      else if(!strcasecmp(s,"Message")) parm_msg(p->g.string,GK_MSG_LEN);
      else if(!strcasecmp(s,"ShootMany")) {
	 p->g.shootnum=parm_num();
	 p->g.shootarc=FIXED_PI/parm_num();
      }
      else if(!strcasecmp(s,"IconPos")) {
	 p->g.xo=parm_num();
	 p->g.yo=parm_num();
      }
      else break;
   };
   unget_token();
};

void phasecomp(void) {
   const char *s;
   ThingPhaseRec *p;
   ThingPhase def;
   ThingPhase *tp=&def;
   /* make new phasetable */
   memset(&def,0,sizeof(ThingPhase));
   def.id=nphasetbls+1;
   def.next=-1;
   def.sound=-1;
   def.wait=1;
   p=phases+(nphasetbls++);
   if(nphasetbls>=LOTS) err("too many phases");
   memset(p,0,sizeof(ThingPhaseRec));
   memset(p->signals,-1,sizeof(short)*NUM_THINGSIGS);
   p->signals[TS_INIT]=0;
   p->tp=(ThingPhase*)malloc(LOTS*sizeof(ThingPhase));
   p->tpname=(char*)malloc(LOTS*NAMELEN);
   if(p->tp==NULL||p->tpname==NULL) 
      err("error allocating lots of thingphases");
   /* one parameter, name */
   s=next_token();
   if(s==NULL||*s=='\n') synerr("name expected after PhaseTable");
   strncpy(p->name,s,NAMELEN-1);
   /* now the phase table */
   while(1) {
      s=next_token();
      if(s==NULL) return;
      else if(*s=='\n');
      else if(!strcasecmp(s,"Default")) tp=&def;
      else if(!strcasecmp(s,"Phase")) {
	 /* add a new phase to the table */
	 tp=p->tp+p->nphases;
	 memcpy(tp,&def,sizeof(ThingPhase));
	 def.id=0;
	 parm_str(p->tpname+p->nphases*NAMELEN,NAMELEN);
	 p->nphases++;
	 s=next_token();
	 if(s!=NULL) tp->spr_phase=*s;
      }
      else if(!strcasecmp(s,"SigDetect")) 
	 p->signals[TS_DETECT]=p->nphases-1;
      else if(!strcasecmp(s,"SigFight"))
	 p->signals[TS_FIGHT]=p->nphases-1;
      else if(!strcasecmp(s,"SigShoot")) 
	 p->signals[TS_SHOOT]=p->nphases-1;
      else if(!strcasecmp(s,"SigSpecial")) 
	 p->signals[TS_SPECIAL]=p->nphases-1;
      else if(!strcasecmp(s,"SigOuch")) 
	 p->signals[TS_OUCH]=p->nphases-1;
      else if(!strcasecmp(s,"SigDie")) 
	 p->signals[TS_DIE]=p->nphases-1;
      else if(!strcasecmp(s,"SigExplode")) 
	 p->signals[TS_EXPLODE]=p->nphases-1;
      else if(!strcasecmp(s,"SigReanimate")) 
	 p->signals[TS_ANIMATE]=p->nphases-1;
      else if(!strcasecmp(s,"SigInit")) 
	 p->signals[TS_INIT]=p->nphases-1;
      else if(!strcasecmp(s,"Glow")) tp->flags|=TPH_GLOW;
      else if(!strcasecmp(s,"Destroy")) tp->flags|=TPH_DESTROY;
      else if(!strcasecmp(s,"NoSigs")) tp->flags|=TPH_NOSIGS;
      else if(!strcasecmp(s,"Strategy")) tp->flags|=TPH_STRATEGY;
      else if(!strcasecmp(s,"HeatSeek")) tp->flags|=TPH_HEATSEEK;
      else if(!strcasecmp(s,"Shoot")) tp->flags|=TPH_SHOOT;
      else if(!strcasecmp(s,"Explode")) tp->flags|=TPH_EXPLODE;
      else if(!strcasecmp(s,"Melee")) tp->flags|=TPH_MELEE;
      else if(!strcasecmp(s,"Spawn2")) tp->flags|=TPH_SPAWN2;
      else if(!strcasecmp(s,"RSpawn2")) tp->flags|=TPH_RSPAWN2;
      else if(!strcasecmp(s,"Idle")) tp->flags|=TPH_IDLE;
      else if(!strcasecmp(s,"Noisy")) tp->flags|=TPH_NOISY;
      else if(!strcasecmp(s,"BFGEffect")) tp->flags|=TPH_BFGEFFECT;
      else if(!strcasecmp(s,"Spawn")) tp->flags|=TPH_SHOOT;
      else if(!strcasecmp(s,"Become")) {
	 tp->flags|=TPH_BECOME;
	 tp->next=0;
      }
      else if(!strcasecmp(s,"Become2")) {
	 tp->flags|=TPH_BECOME2;
	 tp->next=0;
      }
      else if(!strcasecmp(s,"Duration")) tp->wait=parm_num();
      else if(!strcasecmp(s,"Goto")) tp->next=parm_phase(p);
      else if(!strcasecmp(s,"SpritePh")) tp->spr_phase=parm_ch();
      else if(!strcasecmp(s,"Sprite")) parm_str(tp->sprite,5);
      else break;
   };
   p->tp=(ThingPhase*)realloc(p->tp,p->nphases*sizeof(ThingPhase));
   p->tpname=(char *)realloc(p->tpname,p->nphases*NAMELEN);
   unget_token();
};

static void set_phid(ProtoThingRec *p,ThingPhaseRec *tpr) {
   p->pt.phase_id=tpr->tp->id;
   memcpy(p->pt.signals,tpr->signals,sizeof(short)*NUM_THINGSIGS);
};

void protocomp(void) {
   const char *s;
   ProtoThingRec *p;
   ThingPhaseRec *tpr;
   /* make new proto */
   p=protos+(nprotos++);
   if(nprotos>=LOTS) err("too many protos");
   memset(p,0,sizeof(ProtoThingRec));
   p->pt.hits=-1;
   p->pt.artitype=-1;
   p->pt.shootnum=1;
   p->pt.flags|=PT_INF_MASS;
   /* two parameters, name and id */
   s=next_token();
   if(s==NULL||*s=='\n') synerr("name expected after Proto");
   strncpy(p->name,s,NAMELEN-1);
   tpr=find_ph_tbl(p->name);
   if(tpr) set_phid(p,tpr);
   s=next_token();
   if(s==NULL||*s=='\n') 
      p->pt.id=top_assigned_proto_id++;
   else chkuniqueid(p->pt.id=atoi(s));
   /* now info for this proto */
   while(1) {
      s=next_token();
      if(s==NULL) return;
      else if(*s=='\n');
      else if(!strcasecmp(s,"Phases")) set_phid(p,tpr=parm_ph_tbl());
      else if(!strcasecmp(s,"SigDetect")) 
	 p->pt.signals[TS_DETECT]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigFight")) 
	 p->pt.signals[TS_FIGHT]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigShoot")) 
	 p->pt.signals[TS_SHOOT]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigSpecial")) 
	 p->pt.signals[TS_SPECIAL]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigOuch")) 
	 p->pt.signals[TS_OUCH]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigDie")) 
	 p->pt.signals[TS_DIE]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigExplode")) 
	 p->pt.signals[TS_EXPLODE]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigReanimate")) 
	 p->pt.signals[TS_ANIMATE]=parm_phase(tpr);
      else if(!strcasecmp(s,"SigInit")) 
	 p->pt.signals[TS_INIT]=parm_phase(tpr);
      else if(!strcasecmp(s,"Splat")) p->pt.bloodtype=parm_proto();
      else if(!strcasecmp(s,"Shoots")) p->pt.spawn1=parm_proto();
      else if(!strcasecmp(s,"Spawns")) p->pt.spawn1=parm_proto();
      else if(!strcasecmp(s,"Spawn2")) p->pt.spawn2=parm_proto();
      else if(!strcasecmp(s,"Becomes")) p->pt.become1=parm_proto();
      else if(!strcasecmp(s,"Become2")) p->pt.become2=parm_proto();
      else if(!strcasecmp(s,"Sprite")) parm_str(p->pt.sprite,6);
      else if(!strcasecmp(s,"Damage")) p->pt.damage=parm_num();
      else if(!strcasecmp(s,"Mass")) {
	 int i=parm_num();
	 p->pt.flags&=~PT_INF_MASS;
	 if(i) {
	    p->pt.realmass=(FIXED_ONE*i)/100;
	    p->pt.friction=(5*FIXED_ONE)/8;
	 }
	 else {
	    p->pt.realmass=p->pt.friction=FIXED_ONE;
	    p->pt.flags|=PT_CAN_FLY;
	 };
      }
      else if(!strcasecmp(s,"Hits")) p->pt.hits=parm_num();
      else if(!strcasecmp(s,"Shooter")) p->pt.flags|=PT_SHOOTER;
      else if(!strcasecmp(s,"Noisy")) p->pt.flags|=PT_NOISY;
      else if(!strcasecmp(s,"Beastie")) p->pt.flags|=PT_BEASTIE|PT_TARGET;
      else if(!strcasecmp(s,"Explosive")) p->pt.flags|=PT_EXPLOSIVE;
      else if(!strcasecmp(s,"ZCenter")) p->pt.flags|=PT_ZCENTER;
      else if(!strcasecmp(s,"PHANTOM")) p->pt.flags|=PT_PHANTOM;
      else if(!strcasecmp(s,"SKIRTCLIFFS")) p->pt.flags|=PT_SKIRT_CLIFFS;
      else if(!strcasecmp(s,"CANFLY")) p->pt.flags|=PT_CAN_FLY;
      else if(!strcasecmp(s,"Flying")) p->pt.flags|=PT_CAN_FLY;
      else if(!strcasecmp(s,"PINVIS")) p->pt.flags|=PT_PINVIS;
      else if(!strcasecmp(s,"BULLETKLUDGE")) p->pt.flags|=PT_BULLET_KLUDGE;
      else if(!strcasecmp(s,"FASTSHOOT")) p->pt.flags|=PT_FAST_SHOOTER;
      else if(!strcasecmp(s,"HANGING")) p->pt.flags|=PT_HANGING;
      else if(!strcasecmp(s,"Player")) p->pt.flags|=PT_PLAYER;
      else if(!strcasecmp(s,"Target")) p->pt.flags|=PT_TARGET;
      else if(!strcasecmp(s,"TeleSpot")) p->pt.flags|=PT_TELESPOT;
      else if(!strcasecmp(s,"SpawnSpot")) p->pt.flags|=PT_SPAWNSPOT;
      else if(!strcasecmp(s,"ShootMany")) {
	 p->pt.shootnum=parm_num();
	 p->pt.shootarc=FIXED_PI/parm_num();
      }
      else if(!strcasecmp(s,"Size")) {
	 p->pt.radius=parm_num()<<11;
	 p->pt.height=parm_num()<<12;
      }
      else if(!strcasecmp(s,"Gets")) {
	 p->pt.artitype=parm_gett();
	 p->pt.artinum=parm_num();
      }
      else break;
   };
   unget_token();
};

void ptcomp(void) {
   while(1) {
      const char *s=next_token();
      if(s==NULL) break;
      else if(*s=='\n');
      else if(!strcasecmp(s,"Proto")) protocomp();
      else if(!strcasecmp(s,"PhaseTable")) phasecomp();
      else if(!strcasecmp(s,"Gettable")) gettcomp();
      else synerr(NULL);
   };
};

void wrphases(FILE *fout) {
   int i;
   for(i=0;i<nphasetbls;i++)
      fwrite(phases[i].tp,sizeof(ThingPhase),phases[i].nphases,fout);
};
void wrprotos(FILE *fout) {
   int i;
   for(i=0;i<nprotos;i++) 
      fwrite(&protos[i].pt,sizeof(ProtoThing),1,fout);
};
void wrgetts(FILE *fout) {
   int i;
   for(i=0;i<ngetts;i++) 
      fwrite(&getts[i].g,sizeof(Gettable),1,fout);
};

static int ptcmp(const void *p1,const void *p2) {
   if(((const ProtoThingRec *)p1)->pt.id>((const ProtoThingRec *)p2)->pt.id) 
      return 1;
   if(((const ProtoThingRec *)p1)->pt.id<((const ProtoThingRec *)p2)->pt.id) 
      return -1;
   return 0;
};

int main(int argc,char **argv) {
   FILE *fout;
   char foutname[256];
   if(argc!=2) {
      printf("Usage: ptcomp <dir>\n\n"
	     "ptcomp is used to compile .pt files into the PHASES,\n"
	     "PROTOS, and GETTABLE lumps used by DUMB.  It reads .pt\n"
	     "source from the standard input, and writes its output\n"
	     "into files of the form <dir>/PROTOS.lump\n"
	     "\n");
      return 1;
   };
   protos=(ProtoThingRec*)malloc(LOTS*sizeof(ProtoThingRec));
   phases=(ThingPhaseRec*)malloc(LOTS*sizeof(ThingPhaseRec));
   getts=(GettableRec*)malloc(LOTS*sizeof(GettableRec));
   if(protos==NULL||phases==NULL||getts==NULL) {
      printf("Trouble allocating lots of VM for tables\n");
      return 5;
   };
   fin=stdin;
   if(fin==NULL) {
      printf("Error %d opening source file\n",errno);
      return 2;
   };

   printf("ptcomp: compiling...\n");
   ptcomp();
   /*fclose(fin);*/

   printf("ptcomp: %d protos, %d phase tables, %d gettables\n",
	  nprotos,nphasetbls,ngetts);

   printf("ptcomp: sorting...\n");
   qsort(protos,nprotos,sizeof(ProtoThingRec),ptcmp);

   printf("ptcomp: writing...\n");
   fout=fopen(strcat(strcpy(foutname,argv[1]),"/PHASES.lump"),"wb");
   if(fout) {
      wrphases(fout);
      fclose(fout);
   }
   else printf("Error %d opening phase file %s\n",errno,foutname);
   fout=fopen(strcat(strcpy(foutname,argv[1]),"/PROTOS.lump"),"wb");
   if(fout) {
      wrprotos(fout);
      fclose(fout);
   }
   else printf("Error %d opening proto file %s\n",errno,foutname);
   fout=fopen(strcat(strcpy(foutname,argv[1]),"/GETTABLE.lump"),"wb");
   if(fout) {
      wrgetts(fout);
      fclose(fout);
   }
   else printf("Error %d opening gettable file %s\n",errno,foutname);
   return 0;
};
